/* minimization of M(natural)-convex function
 *  This is an implementation of scaling algorithm
 *  (improved steepest descent method).
 *  The calculation cost is O(dimension) per step.
 *  cf. S. Moriguchi, K. Murota, and A. Shioura,
 *      Scaling Algorithms for M-convex Function Minimization,
 *      Trans. IEICE, Fundamentals of Electronics,
 *      Communications and Computer Sciences, {\bf E85}\, (2002), 922--929.
 */


#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <assert.h>
#include "odicon.h"

static void status_print2(int dim, double func(int,int*), int x[],
                          int lo[], int up[], int width) {
#ifdef DEBUG
    printf("f(x) = %3g,  x = ", func(dim, x));
    ivector_print(dim, x);
    printf(",  s = %d,  bound = ", width);
    ivector_print(dim, lo);
    ivector_print(dim, up);
    printf("\n");
#endif
}

/* returns index (for plus) if improvment was found. */
/* returns 'minus' if no improvment was found. */
static int best_improve(int dim, double func(int,int*), int x[],
                        int minus, int up[], int width, int natural) {
    int i, min_i = minus;
    double f, min_f;

    iteration++;
    min_f = func(dim, x);  oracle_call++;
    x[minus] -= width;

    for (i=0; i<dim+natural; i++) {
        if (i == minus) continue;
        x[i] += width;
        if (i == dim || x[i] <= up[i]) {
            f = func(dim, x);  oracle_call++;
            if (f < min_f) {  min_f = f;  min_i = i;  improve_call++;  }
        }
        x[i] -= width;
    }

    x[minus] += width;
    return min_i;
}

/* 'lo' will be destroied */
static void steepest_descent(int dim, double func(int,int*), int x[],
                             int lo[], int up[], int width, int natural) {
    int minus, plus; /* the indexes of 'x[]'. */
    int improved;
    int init_iteration = iteration;

    status_print2(dim, func, x, lo, up, width);
    do {
        improved = FALSE;
        for (minus=0; minus<dim+natural; minus++) {
            if (minus == dim || lo[minus] <= x[minus] - width) {
                plus = best_improve(dim, func, x, minus, up, width, natural);
                if (plus != minus) {
                    improved = TRUE;
                    x[plus]  += width;
                    x[minus] -= width;
                }
                lo[plus] = x[plus];
                status_print2(dim, func, x, lo, up, width);
            }
        }
    } while (improved);
    if (width > 1 && iteration - init_iteration > dim * dim) {
        printf("#iteration = %d > dim^2\n", iteration - init_iteration);
    }
}


/* returns the minimum value of func(x)    */
/* INPUT  func   : pointer to the function */
/* INPUT  init[] : the initial solution    */
/* OUTPUT init[] : the minimized solution  */
static double minimize(int dim, double func(int,int*), int init[],
                       int lower[], int upper[], int natural) {
    int i, width = INT_MAX, radius;
    int *x, *lo, *up;

    counter_reset();
    if (!is_feasible(dim, func, init)) return PLUS_INFINITY;
    status_print(dim, func, init);

    for (i=0; i<dim; i++) {
        assert(lower[i] <= init[i]);
        assert(init[i] <= upper[i]);
        if (width > upper[i] - lower[i]) {
            width = upper[i] - lower[i];
        }
    }
    width /= dim;

    assert(natural == 0 || natural == 1);
    x  = ivector_new(dim + natural);
    lo = ivector_new(dim + natural);
    up = ivector_new(dim + natural);
    ivector_copy(dim, x,  init);
    ivector_copy(dim, lo, lower);
    ivector_copy(dim, up, upper);

    do {
        /* one step */
        steepest_descent(dim, func, x, lo, up, width, natural);
        ivector_copy(dim, lo, lower);

        /* update boundary */
        radius = (dim - 1 + natural) * (width - 1);
        for (i=0; i<dim; i++) {
            if (lo[i] < x[i] - radius) {
                lo[i] = x[i] - radius;
            }
            if (up[i] > x[i] + radius) {
                up[i] = x[i] + radius;
            }
        }
        width /= 2 * sqrt(dim);  /* imploved by N.T */
    } while (width > 1);

    steepest_descent(dim, func, x, lo, up, 1, natural);
    status_print(dim, func, x);

    ivector_copy(dim, init, x);
    free(x);
    free(lo);
    free(up);

    printf("\n");
    counter_print();
    minimizer_print(dim, func, init);
    return func(dim, init);
}


double mconv_minimize_scaling(int dim, double func(int,int*), int init[],
                              int lower[], int upper[]) {
    return minimize(dim, func, init, lower, upper, 0);
}

double mgconv_minimize_scaling(int dim, double func(int,int*), int init[],
                              int lower[], int upper[]) {
    return minimize(dim, func, init, lower, upper, 1);
}
